import contextlib
import textwrap
import secrets
import subprocess
import pathlib
import functools
import getpass
import sys
import datetime

import pytest
import portend
import requests
import jaraco.envs
import munch

from twine import settings


@pytest.fixture()
def pypirc(tmpdir):
    return tmpdir / ".pypirc"


@pytest.fixture()
def make_settings(pypirc):
    """Returns a factory function for settings.Settings with defaults."""

    default_pypirc = """
        [pypi]
        username:foo
        password:bar
    """

    def _settings(pypirc_text=default_pypirc, **settings_kwargs):
        pypirc.write(textwrap.dedent(pypirc_text))

        settings_kwargs.setdefault('sign_with', None)
        settings_kwargs.setdefault('config_file', str(pypirc))

        return settings.Settings(**settings_kwargs)

    return _settings


class DevPiEnv(jaraco.envs.ToxEnv):
    """
    Run devpi using tox:testenv:devpi.
    """
    name = 'devpi'
    username = 'foober'

    def create(self, root, password):
        super().create()
        self.base = root
        self.password = password
        self.port = portend.find_available_local_port()
        cmd = [
            self.exe('devpi-init'),
            '--serverdir', str(root),
            '--root-passwd', password,
        ]
        subprocess.run(cmd, check=True)

    @property
    def url(self):
        return f'http://localhost:{self.port}/'

    @property
    def repo(self):
        return f'{self.url}/{self.username}/dev/'

    def ready(self):
        with contextlib.suppress(Exception):
            return requests.get(self.url)

    def init(self):
        run = functools.partial(subprocess.run, check=True)
        client_dir = self.base / 'client'
        devpi_client = [
            self.exe('devpi'),
            '--clientdir', str(client_dir),
        ]
        run(devpi_client + ['use', self.url + 'root/pypi/'])
        run(devpi_client + [
            'user', '--create', self.username, f'password={self.password}'])
        run(devpi_client + [
            'login', self.username, '--password', self.password])
        run(devpi_client + ['index', '-c', 'dev'])


@pytest.fixture(scope='session')
def devpi_server(request, watcher_getter, tmp_path_factory):
    env = DevPiEnv()
    password = secrets.token_urlsafe()
    root = tmp_path_factory.mktemp('devpi')
    env.create(root, password)
    proc = watcher_getter(
        name=str(env.exe('devpi-server')),
        arguments=['--port', str(env.port), '--serverdir', str(root)],
        checker=env.ready,
        # Needed for the correct execution order of finalizers
        request=request,
    )
    env.init()
    username = env.username
    url = env.repo
    return munch.Munch.fromDict(locals())


dist_names = [
    'twine-1.5.0.tar.gz',
    'twine-1.5.0-py2.py3-none-any.whl',
    'twine-1.6.5.tar.gz',
    'twine-1.6.5-py2.py3-none-any.whl',
]


@pytest.fixture(params=dist_names)
def uploadable_dist(request):
    return pathlib.Path(__file__).parent / 'fixtures' / request.param


@pytest.fixture
def entered_password(monkeypatch):
    monkeypatch.setattr(getpass, 'getpass', lambda prompt: 'entered pw')


@pytest.fixture
def sampleproject_dist(tmp_path_factory):
    checkout = tmp_path_factory.mktemp('sampleproject', numbered=False)
    subprocess.run([
        'git',
        'clone', 'https://github.com/pypa/sampleproject',
        str(checkout),
    ])
    with (checkout / 'setup.py').open('r+') as setup:
        orig = setup.read()
        sub = orig.replace(
            "name='sampleproject'", "name='twine-sampleproject'")
        assert orig != sub
        setup.seek(0)
        setup.write(sub)
    tag = datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')
    subprocess.run([
        sys.executable,
        'setup.py',
        'egg_info', '--tag-build', f'post{tag}',
        'sdist',
    ],
        cwd=str(checkout),
    )
    dist, = checkout.joinpath('dist').glob('*')
    return dist


class PypiserverEnv(jaraco.envs.ToxEnv):
    """
    Run pypiserver using tox:testenv:pypiserver.
    """
    name = 'pypiserver'

    @property
    @functools.lru_cache()
    def port(self):
        return portend.find_available_local_port()

    @property
    def url(self):
        return f'http://localhost:{self.port}/'

    def ready(self):
        with contextlib.suppress(Exception):
            return requests.get(self.url)


@pytest.fixture(scope='session')
def pypiserver_instance(request, watcher_getter, tmp_path_factory):
    env = PypiserverEnv()
    env.create()
    proc = watcher_getter(
        name=str(env.exe('pypi-server')),
        arguments=[
            '--port', str(env.port),
            # allow anonymous uploads
            '-P', '.', '-a', '.',
            tmp_path_factory.mktemp('packages'),
        ],
        checker=env.ready,
        # Needed for the correct execution order of finalizers
        request=request,
    )
    url = env.url
    return munch.Munch.fromDict(locals())



# Copyright 2014 Ian Cordasco
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pretend
import pytest
from requests.exceptions import HTTPError

from twine.commands import upload
from twine import package, cli, exceptions
import twine

import helpers

SDIST_FIXTURE = 'tests/fixtures/twine-1.5.0.tar.gz'
WHEEL_FIXTURE = 'tests/fixtures/twine-1.5.0-py2.py3-none-any.whl'
RELEASE_URL = 'https://pypi.org/project/twine/1.5.0/'
NEW_SDIST_FIXTURE = 'tests/fixtures/twine-1.6.5.tar.gz'
NEW_WHEEL_FIXTURE = 'tests/fixtures/twine-1.6.5-py2.py3-none-any.whl'
NEW_RELEASE_URL = 'https://pypi.org/project/twine/1.6.5/'


def test_successful_upload(make_settings, capsys):
    upload_settings = make_settings()

    stub_response = pretend.stub(
        is_redirect=False,
        status_code=201,
        raise_for_status=lambda: None
    )

    stub_repository = pretend.stub(
        upload=lambda package: stub_response,
        close=lambda: None,
        release_urls=lambda packages: {RELEASE_URL, NEW_RELEASE_URL}
    )

    upload_settings.create_repository = lambda: stub_repository

    result = upload.upload(upload_settings, [
        WHEEL_FIXTURE, SDIST_FIXTURE, NEW_SDIST_FIXTURE, NEW_WHEEL_FIXTURE
    ])

    # A truthy result means the upload failed
    assert result is None

    captured = capsys.readouterr()
    assert captured.out.count(RELEASE_URL) == 1
    assert captured.out.count(NEW_RELEASE_URL) == 1


@pytest.mark.parametrize('verbose', [False, True])
def test_exception_for_http_status(verbose, make_settings, capsys):
    upload_settings = make_settings()
    upload_settings.verbose = verbose

    stub_response = pretend.stub(
        is_redirect=False,
        status_code=403,
        text="Invalid or non-existent authentication information",
        raise_for_status=pretend.raiser(HTTPError)
    )

    stub_repository = pretend.stub(
        upload=lambda package: stub_response,
        close=lambda: None,
    )

    upload_settings.create_repository = lambda: stub_repository

    with pytest.raises(HTTPError):
        upload.upload(upload_settings, [WHEEL_FIXTURE])

    captured = capsys.readouterr()
    assert RELEASE_URL not in captured.out

    if verbose:
        assert stub_response.text in captured.out
        assert '--verbose' not in captured.out
    else:
        assert stub_response.text not in captured.out
        assert '--verbose' in captured.out


def test_get_config_old_format(make_settings, pypirc):
    try:
        make_settings("""
            [server-login]
            username:foo
            password:bar
        """)
    except KeyError as err:
        assert all(text in err.args[0] for text in [
            "'pypi'",
            "--repository-url",
            pypirc,
            "https://docs.python.org/",
        ])


def test_deprecated_repo(make_settings):
    with pytest.raises(exceptions.UploadToDeprecatedPyPIDetected) as err:
        upload_settings = make_settings("""
            [pypi]
            repository: https://pypi.python.org/pypi/
            username:foo
            password:bar
        """)

        upload.upload(upload_settings, [WHEEL_FIXTURE])

    assert all(text in err.value.args[0] for text in [
        "https://pypi.python.org/pypi/",
        "https://upload.pypi.org/legacy/",
        "https://test.pypi.org/legacy/",
        "https://packaging.python.org/",
    ])


def test_exception_for_redirect(make_settings):
    upload_settings = make_settings("""
        [pypi]
        repository: https://test.pypi.org/legacy
        username:foo
        password:bar
    """)

    stub_response = pretend.stub(
        is_redirect=True,
        status_code=301,
        headers={'location': 'https://test.pypi.org/legacy/'}
    )

    stub_repository = pretend.stub(
        upload=lambda package: stub_response,
        close=lambda: None
    )

    upload_settings.create_repository = lambda: stub_repository

    with pytest.raises(exceptions.RedirectDetected) as err:
        upload.upload(upload_settings, [WHEEL_FIXTURE])

    assert "https://test.pypi.org/legacy/" in err.value.args[0]


def test_prints_skip_message_for_uploaded_package(make_settings, capsys):
    upload_settings = make_settings(skip_existing=True)

    stub_repository = pretend.stub(
        # Short-circuit the upload, so no need for a stub response
        package_is_uploaded=lambda package: True,
        release_urls=lambda packages: {},
        close=lambda: None
    )

    upload_settings.create_repository = lambda: stub_repository

    result = upload.upload(upload_settings, [WHEEL_FIXTURE])

    # A truthy result means the upload failed
    assert result is None

    captured = capsys.readouterr()
    assert "Skipping twine-1.5.0-py2.py3-none-any.whl" in captured.out
    assert RELEASE_URL not in captured.out


def test_prints_skip_message_for_response(make_settings, capsys):
    upload_settings = make_settings(skip_existing=True)

    stub_response = pretend.stub(
        is_redirect=False,
        status_code=409,
    )

    stub_repository = pretend.stub(
        # Do the upload, triggering the error response
        package_is_uploaded=lambda package: False,
        release_urls=lambda packages: {},
        upload=lambda package: stub_response,
        close=lambda: None
    )

    upload_settings.create_repository = lambda: stub_repository

    result = upload.upload(upload_settings, [WHEEL_FIXTURE])

    # A truthy result means the upload failed
    assert result is None

    captured = capsys.readouterr()
    assert "Skipping twine-1.5.0-py2.py3-none-any.whl" in captured.out
    assert RELEASE_URL not in captured.out


@pytest.mark.parametrize('response_kwargs', [
    pytest.param(
        dict(status_code=400, reason=(
            'A file named "twine-1.5.0-py2.py3-none-any.whl" already '
            'exists for twine-1.5.0.'
        )),
        id='pypi'
    ),
    pytest.param(
        dict(status_code=400, reason=(
            'Repository does not allow updating assets: pypi for url: '
            'http://www.foo.bar'
        )),
        id='nexus'
    ),
    pytest.param(
        dict(status_code=409, reason=(
            'A file named "twine-1.5.0-py2.py3-none-any.whl" already '
            'exists for twine-1.5.0.'
        )),
        id='pypiserver'
    ),
    pytest.param(
        dict(status_code=403, text=(
            "Not enough permissions to overwrite artifact "
            "'pypi-local:twine/1.5.0/twine-1.5.0-py2.py3-none-any.whl'"
            "(user 'twine-deployer' needs DELETE permission)."
        )),
        id='artifactory_old'
    ),
    pytest.param(
        dict(status_code=403, text=(
            "Not enough permissions to delete/overwrite artifact "
            "'pypi-local:twine/1.5.0/twine-1.5.0-py2.py3-none-any.whl'"
            "(user 'twine-deployer' needs DELETE permission)."
        )),
        id='artifactory_new'
    ),
])
def test_skip_existing_skips_files_on_repository(response_kwargs):
    assert upload.skip_upload(
        response=pretend.stub(**response_kwargs),
        skip_existing=True,
        package=package.PackageFile.from_filename(WHEEL_FIXTURE, None),
    )


@pytest.mark.parametrize('response_kwargs', [
    pytest.param(
        dict(status_code=400, reason='Invalid credentials'),
        id='wrong_reason'
    ),
    pytest.param(
        dict(status_code=404),
        id='wrong_code'
    ),
])
def test_skip_upload_doesnt_match(response_kwargs):
    assert not upload.skip_upload(
        response=pretend.stub(**response_kwargs),
        skip_existing=True,
        package=package.PackageFile.from_filename(WHEEL_FIXTURE, None)
    )


def test_skip_upload_respects_skip_existing():
    assert not upload.skip_upload(
        response=pretend.stub(),
        skip_existing=False,
        package=package.PackageFile.from_filename(WHEEL_FIXTURE, None),
    )


def test_values_from_env(monkeypatch):
    def none_upload(*args, **settings_kwargs):
        pass

    replaced_upload = pretend.call_recorder(none_upload)
    monkeypatch.setattr(twine.commands.upload, "upload", replaced_upload)
    testenv = {"TWINE_USERNAME": "pypiuser",
               "TWINE_PASSWORD": "pypipassword",
               "TWINE_CERT": "/foo/bar.crt"}
    with helpers.set_env(**testenv):
        cli.dispatch(["upload", "path/to/file"])
    upload_settings = replaced_upload.calls[0].args[0]
    assert "pypipassword" == upload_settings.password
    assert "pypiuser" == upload_settings.username
    assert "/foo/bar.crt" == upload_settings.cacert


@pytest.mark.parametrize('repo_url', [
    "https://upload.pypi.org/",
    "https://test.pypi.org/",
    "https://pypi.org/"
])
def test_check_status_code_for_wrong_repo_url(repo_url, make_settings):
    upload_settings = make_settings()

    # override defaults to use incorrect URL
    upload_settings.repository_config['repository'] = repo_url

    with pytest.raises(twine.exceptions.InvalidPyPIUploadURL):
        upload.upload(upload_settings, [
            WHEEL_FIXTURE, SDIST_FIXTURE, NEW_SDIST_FIXTURE, NEW_WHEEL_FIXTURE
        ])



# Copyright 2015 Ian Cordasco
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from twine import package, exceptions

import pretend
import pytest


def test_sign_file(monkeypatch):
    replaced_check_call = pretend.call_recorder(lambda args: None)
    monkeypatch.setattr(package.subprocess, 'check_call', replaced_check_call)
    filename = 'tests/fixtures/deprecated-pypirc'

    pkg = package.PackageFile(
        filename=filename,
        comment=None,
        metadata=pretend.stub(name="deprecated-pypirc"),
        python_version=None,
        filetype=None
    )
    try:
        pkg.sign('gpg2', None)
    except IOError:
        pass
    args = ('gpg2', '--detach-sign', '-a', filename)
    assert replaced_check_call.calls == [pretend.call(args)]


def test_sign_file_with_identity(monkeypatch):
    replaced_check_call = pretend.call_recorder(lambda args: None)
    monkeypatch.setattr(package.subprocess, 'check_call', replaced_check_call)
    filename = 'tests/fixtures/deprecated-pypirc'

    pkg = package.PackageFile(
        filename=filename,
        comment=None,
        metadata=pretend.stub(name="deprecated-pypirc"),
        python_version=None,
        filetype=None
    )
    try:
        pkg.sign('gpg', 'identity')
    except IOError:
        pass
    args = ('gpg', '--detach-sign', '--local-user', 'identity', '-a', filename)
    assert replaced_check_call.calls == [pretend.call(args)]


def test_run_gpg_raises_exception_if_no_gpgs(monkeypatch):
    replaced_check_call = pretend.raiser(FileNotFoundError('not found'))
    monkeypatch.setattr(package.subprocess, 'check_call', replaced_check_call)
    gpg_args = ('gpg', '--detach-sign', '-a', 'pypircfile')

    with pytest.raises(exceptions.InvalidSigningExecutable) as err:
        package.PackageFile.run_gpg(gpg_args)

    assert 'executables not available' in err.value.args[0]


def test_run_gpg_raises_exception_if_not_using_gpg(monkeypatch):
    replaced_check_call = pretend.raiser(FileNotFoundError('not found'))
    monkeypatch.setattr(package.subprocess, 'check_call', replaced_check_call)
    gpg_args = ('not_gpg', '--detach-sign', '-a', 'pypircfile')

    with pytest.raises(exceptions.InvalidSigningExecutable) as err:
        package.PackageFile.run_gpg(gpg_args)

    assert 'not_gpg executable not available' in err.value.args[0]


def test_run_gpg_falls_back_to_gpg2(monkeypatch):

    def check_call(arg_list):
        if arg_list[0] == 'gpg':
            raise FileNotFoundError('gpg not found')

    replaced_check_call = pretend.call_recorder(check_call)
    monkeypatch.setattr(package.subprocess, 'check_call', replaced_check_call)
    gpg_args = ('gpg', '--detach-sign', '-a', 'pypircfile')

    package.PackageFile.run_gpg(gpg_args)

    gpg2_args = replaced_check_call.calls[1].args
    assert gpg2_args[0][0] == 'gpg2'


def test_package_signed_name_is_correct():
    filename = 'tests/fixtures/deprecated-pypirc'

    pkg = package.PackageFile(
        filename=filename,
        comment=None,
        metadata=pretend.stub(name="deprecated-pypirc"),
        python_version=None,
        filetype=None
    )

    assert pkg.signed_basefilename == "deprecated-pypirc.asc"
    assert pkg.signed_filename == (filename + '.asc')


@pytest.mark.parametrize('gpg_signature', [
    (None),
    (pretend.stub()),
])
def test_metadata_dictionary(gpg_signature):
    meta = pretend.stub(
        name='whatever',
        version=pretend.stub(),
        metadata_version=pretend.stub(),
        summary=pretend.stub(),
        home_page=pretend.stub(),
        author=pretend.stub(),
        author_email=pretend.stub(),
        maintainer=pretend.stub(),
        maintainer_email=pretend.stub(),
        license=pretend.stub(),
        description=pretend.stub(),
        keywords=pretend.stub(),
        platforms=pretend.stub(),
        classifiers=pretend.stub(),
        download_url=pretend.stub(),
        supported_platforms=pretend.stub(),
        provides=pretend.stub(),
        requires=pretend.stub(),
        obsoletes=pretend.stub(),
        project_urls=pretend.stub(),
        provides_dist=pretend.stub(),
        obsoletes_dist=pretend.stub(),
        requires_dist=pretend.stub(),
        requires_external=pretend.stub(),
        requires_python=pretend.stub(),
        provides_extras=pretend.stub(),
        description_content_type=pretend.stub(),
    )

    pkg = package.PackageFile(
        filename='tests/fixtures/twine-1.5.0-py2.py3-none-any.whl',
        comment=pretend.stub(),
        metadata=meta,
        python_version=pretend.stub(),
        filetype=pretend.stub(),
    )
    pkg.gpg_signature = gpg_signature

    result = pkg.metadata_dictionary()

    # identify release
    assert result['name'] == pkg.safe_name
    assert result['version'] == meta.version

    # file content
    assert result['filetype'] == pkg.filetype
    assert result['pyversion'] == pkg.python_version

    # additional meta-data
    assert result['metadata_version'] == meta.metadata_version
    assert result['summary'] == meta.summary
    assert result['home_page'] == meta.home_page
    assert result['author'] == meta.author
    assert result['author_email'] == meta.author_email
    assert result['maintainer'] == meta.maintainer
    assert result['maintainer_email'] == meta.maintainer_email
    assert result['license'] == meta.license
    assert result['description'] == meta.description
    assert result['keywords'] == meta.keywords
    assert result['platform'] == meta.platforms
    assert result['classifiers'] == meta.classifiers
    assert result['download_url'] == meta.download_url
    assert result['supported_platform'] == meta.supported_platforms
    assert result['comment'] == pkg.comment

    # PEP 314
    assert result['provides'] == meta.provides
    assert result['requires'] == meta.requires
    assert result['obsoletes'] == meta.obsoletes

    # Metadata 1.2
    assert result['project_urls'] == meta.project_urls
    assert result['provides_dist'] == meta.provides_dist
    assert result['obsoletes_dist'] == meta.obsoletes_dist
    assert result['requires_dist'] == meta.requires_dist
    assert result['requires_external'] == meta.requires_external
    assert result['requires_python'] == meta.requires_python

    # Metadata 2.1
    assert result['provides_extras'] == meta.provides_extras
    assert result['description_content_type'] == meta.description_content_type

    # GPG signature
    assert result.get('gpg_signature') == gpg_signature


TWINE_1_5_0_WHEEL_HEXDIGEST = package.Hexdigest(
    '1919f967e990bee7413e2a4bc35fd5d1',
    'd86b0f33f0c7df49e888b11c43b417da5520cbdbce9f20618b1494b600061e67',
    'b657a4148d05bd0098c1d6d8cc4e14e766dbe93c3a5ab6723b969da27a87bac0',
)


def test_hash_manager():
    """Verify our HashManager works."""
    filename = 'tests/fixtures/twine-1.5.0-py2.py3-none-any.whl'
    hasher = package.HashManager(filename)
    hasher.hash()
    assert hasher.hexdigest() == TWINE_1_5_0_WHEEL_HEXDIGEST


def test_fips_hash_manager(monkeypatch):
    """Verify the behaviour if hashlib is using FIPS mode."""
    replaced_md5 = pretend.raiser(ValueError('fipsmode'))
    monkeypatch.setattr(package.hashlib, 'md5', replaced_md5)

    filename = 'tests/fixtures/twine-1.5.0-py2.py3-none-any.whl'
    hasher = package.HashManager(filename)
    hasher.hash()
    hashes = TWINE_1_5_0_WHEEL_HEXDIGEST._replace(md5=None)
    assert hasher.hexdigest() == hashes


def test_no_blake2_hash_manager(monkeypatch):
    """Verify the behaviour with missing blake2."""
    monkeypatch.setattr(package, 'blake2b', None)

    filename = 'tests/fixtures/twine-1.5.0-py2.py3-none-any.whl'
    hasher = package.HashManager(filename)
    hasher.hash()
    hashes = TWINE_1_5_0_WHEEL_HEXDIGEST._replace(blake2=None)
    assert hasher.hexdigest() == hashes


def test_pkginfo_returns_no_metadata(monkeypatch):
    """
    Fail gracefully if pkginfo can't interpret the metadata (possibly due to
    seeing a version number it doesn't support yet) and gives us back an
    'empty' object with no metadata
    """

    def EmptyDist(filename):
        return pretend.stub(name=None, version=None)

    monkeypatch.setattr(package, "DIST_TYPES", {"bdist_wheel": EmptyDist})
    filename = 'tests/fixtures/twine-1.5.0-py2.py3-none-any.whl'

    with pytest.raises(exceptions.InvalidDistribution) as err:
        package.PackageFile.from_filename(filename, comment=None)

    assert 'Invalid distribution metadata' in err.value.args[0]



# Copyright 2013 Donald Stufft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import textwrap

import pytest
import pretend

from twine import exceptions, utils

import helpers


def test_get_config(tmpdir):
    pypirc = os.path.join(str(tmpdir), ".pypirc")

    with open(pypirc, "w") as fp:
        fp.write(textwrap.dedent("""
            [distutils]
            index-servers = pypi

            [pypi]
            username = testuser
            password = testpassword
        """))

    assert utils.get_config(pypirc) == {
        "pypi": {
            "repository": utils.DEFAULT_REPOSITORY,
            "username": "testuser",
            "password": "testpassword",
        },
    }


def test_get_config_no_distutils(tmpdir):
    """
    Even if the user hasn't set PyPI has an index server
    in 'index-servers', default to uploading to PyPI.
    """
    pypirc = os.path.join(str(tmpdir), ".pypirc")

    with open(pypirc, "w") as fp:
        fp.write(textwrap.dedent("""
            [pypi]
            username = testuser
            password = testpassword
        """))

    assert utils.get_config(pypirc) == {
        "pypi": {
            "repository": utils.DEFAULT_REPOSITORY,
            "username": "testuser",
            "password": "testpassword",
        },
        "testpypi": {
            "repository": utils.TEST_REPOSITORY,
            "username": None,
            "password": None,
        },
    }


def test_get_config_no_section(tmpdir):
    pypirc = os.path.join(str(tmpdir), ".pypirc")

    with open(pypirc, "w") as fp:
        fp.write(textwrap.dedent("""
            [distutils]
            index-servers = pypi foo

            [pypi]
            username = testuser
            password = testpassword
        """))

    assert utils.get_config(pypirc) == {
        "pypi": {
            "repository": utils.DEFAULT_REPOSITORY,
            "username": "testuser",
            "password": "testpassword",
        },
    }


def test_get_config_override_pypi_url(tmpdir):
    pypirc = os.path.join(str(tmpdir), ".pypirc")

    with open(pypirc, "w") as fp:
        fp.write(textwrap.dedent("""
            [pypi]
            repository = http://pypiproxy
        """))

    assert utils.get_config(pypirc)['pypi']['repository'] == 'http://pypiproxy'


def test_get_config_missing(tmpdir):
    pypirc = os.path.join(str(tmpdir), ".pypirc")

    assert utils.get_config(pypirc) == {
        "pypi": {
            "repository": utils.DEFAULT_REPOSITORY,
            "username": None,
            "password": None,
        },
        "testpypi": {
            "repository": utils.TEST_REPOSITORY,
            "username": None,
            "password": None
        },
    }


def test_empty_userpass(tmpdir):
    """
    Empty username and password may be supplied to suppress
    prompts. See #426.
    """
    pypirc = os.path.join(str(tmpdir), ".pypirc")

    with open(pypirc, "w") as fp:
        fp.write(textwrap.dedent("""
            [pypi]
            username=
            password=
        """))

    config = utils.get_config(pypirc)
    pypi = config['pypi']

    assert pypi['username'] == pypi['password'] == ''


def test_get_repository_config_missing(tmpdir):
    pypirc = os.path.join(str(tmpdir), ".pypirc")

    repository_url = "https://notexisting.python.org/pypi"
    exp = {
        "repository": repository_url,
        "username": None,
        "password": None,
    }
    assert (utils.get_repository_from_config(pypirc, 'foo', repository_url) ==
            exp)
    assert (utils.get_repository_from_config(pypirc, 'pypi', repository_url) ==
            exp)
    exp = {
            "repository": utils.DEFAULT_REPOSITORY,
            "username": None,
            "password": None,
        }
    assert utils.get_repository_from_config(pypirc, "pypi") == exp


def test_get_config_deprecated_pypirc():
    tests_dir = os.path.dirname(os.path.abspath(__file__))
    deprecated_pypirc_path = os.path.join(tests_dir, 'fixtures',
                                          'deprecated-pypirc')

    assert utils.get_config(deprecated_pypirc_path) == {
        "pypi": {
            "repository": utils.DEFAULT_REPOSITORY,
            "username": "testusername",
            "password": "testpassword",
        },
        "testpypi": {
            "repository": utils.TEST_REPOSITORY,
            "username": "testusername",
            "password": "testpassword",
        },
    }


@pytest.mark.parametrize(
    ('cli_value', 'config', 'key', 'strategy', 'expected'),
    (
        ('cli', {}, 'key', lambda: 'fallback', 'cli'),
        (None, {'key': 'value'}, 'key', lambda: 'fallback', 'value'),
        (None, {}, 'key', lambda: 'fallback', 'fallback'),
    ),
)
def test_get_userpass_value(cli_value, config, key, strategy, expected):
    ret = utils.get_userpass_value(cli_value, config, key, strategy)
    assert ret == expected


@pytest.mark.parametrize(
    ('env_name', 'default', 'environ', 'expected'),
    [
        ('MY_PASSWORD', None, {}, None),
        ('MY_PASSWORD', None, {'MY_PASSWORD': 'foo'}, 'foo'),
        ('URL', 'https://example.org', {}, 'https://example.org'),
        ('URL', 'https://example.org', {'URL': 'https://pypi.org'},
            'https://pypi.org'),
    ],
)
def test_default_to_environment_action(env_name, default, environ, expected):
    option_strings = ('-x', '--example')
    dest = 'example'
    with helpers.set_env(**environ):
        action = utils.EnvironmentDefault(
            env=env_name,
            default=default,
            option_strings=option_strings,
            dest=dest,
        )
    assert action.env == env_name
    assert action.default == expected


@pytest.mark.parametrize('repo_url', [
    "https://pypi.python.org",
    "https://testpypi.python.org"
])
def test_check_status_code_for_deprecated_pypi_url(repo_url):
    response = pretend.stub(
        status_code=410,
        url=repo_url
    )

    # value of Verbose doesn't matter for this check
    with pytest.raises(exceptions.UploadToDeprecatedPyPIDetected):
        utils.check_status_code(response, False)



import warnings
import getpass
import functools
from typing import Optional, Callable

import keyring

from . import utils
from . import exceptions


class CredentialInput:

    def __init__(self, username: str = None, password: str = None):
        self.username = username
        self.password = password


class Resolver:
    def __init__(self, config: utils.RepositoryConfig, input: CredentialInput):
        self.config = config
        self.input = input

    @classmethod
    def choose(cls, interactive):
        return cls if interactive else Private

    @property  # type: ignore  # https://github.com/python/mypy/issues/1362
    @functools.lru_cache()
    def username(self) -> Optional[str]:
        return utils.get_userpass_value(
            self.input.username,
            self.config,
            key='username',
            prompt_strategy=self.username_from_keyring_or_prompt,
        )

    @property  # type: ignore  # https://github.com/python/mypy/issues/1362
    @functools.lru_cache()
    def password(self) -> Optional[str]:
        return utils.get_userpass_value(
            self.input.password,
            self.config,
            key='password',
            prompt_strategy=self.password_from_keyring_or_prompt,
        )

    @property
    def system(self) -> Optional[str]:
        return self.config['repository']

    def get_username_from_keyring(self) -> Optional[str]:
        try:
            creds = keyring.get_credential(self.system, None)
            if creds:
                return creds.username
        except AttributeError:
            # To support keyring prior to 15.2
            pass
        except Exception as exc:
            warnings.warn(str(exc))
        return None

    def get_password_from_keyring(self) -> Optional[str]:
        try:
            return keyring.get_password(self.system, self.username)
        except Exception as exc:
            warnings.warn(str(exc))
        return None

    def username_from_keyring_or_prompt(self) -> str:
        return (
            self.get_username_from_keyring()
            or self.prompt('username', input)
        )

    def password_from_keyring_or_prompt(self) -> str:
        return (
            self.get_password_from_keyring()
            or self.prompt('password', getpass.getpass)
        )

    def prompt(self, what: str, how: Callable) -> str:
        return how(f"Enter your {what}: ")


class Private(Resolver):
    def prompt(self, what: str, how: Optional[Callable] = None) -> str:
        raise exceptions.NonInteractive(f"Credential not found for {what}.")



# Copyright 2018 Donald Stufft and individual contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__all__ = (
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
)

__copyright__ = "Copyright 2019 Donald Stufft and individual contributors"

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata


metadata = importlib_metadata.metadata('twine')


__title__ = metadata['name']
__summary__ = metadata['summary']
__uri__ = metadata['home-page']
__version__ = metadata['version']
__author__ = metadata['author']
__email__ = metadata['author-email']
__license__ = metadata['license']



"""Tests for the Settings class and module."""
# Copyright 2018 Ian Stapleton Cordasco
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os.path
import textwrap
import argparse

from twine import exceptions
from twine import settings

import pytest


def test_settings_takes_no_positional_arguments():
    """Verify that the Settings initialization is kw-only."""
    with pytest.raises(TypeError):
        settings.Settings('a', 'b', 'c')


def test_settings_transforms_config(tmpdir):
    """Verify that the settings object transforms the passed in options."""
    pypirc = os.path.join(str(tmpdir), ".pypirc")

    with open(pypirc, "w") as fp:
        fp.write(textwrap.dedent("""
            [pypi]
            repository: https://upload.pypi.org/legacy/
            username:username
            password:password
        """))
    s = settings.Settings(config_file=pypirc)
    assert (s.repository_config['repository'] ==
            'https://upload.pypi.org/legacy/')
    assert s.sign is False
    assert s.sign_with == 'gpg'
    assert s.identity is None
    assert s.username == 'username'
    assert s.password == 'password'
    assert s.cacert is None
    assert s.client_cert is None
    assert s.disable_progress_bar is False


def test_identity_requires_sign():
    """Verify that if a user passes identity, we require sign=True."""
    with pytest.raises(exceptions.InvalidSigningConfiguration):
        settings.Settings(sign=False, identity='fakeid')


def test_password_not_required_if_client_cert(entered_password):
    """Verify that if client_cert is provided then a password is not."""
    test_client_cert = '/random/path'
    settings_obj = settings.Settings(
        username='fakeuser', client_cert=test_client_cert)
    assert not settings_obj.password
    assert settings_obj.client_cert == test_client_cert


@pytest.mark.parametrize('client_cert', [None, ""])
def test_password_is_required_if_no_client_cert(client_cert, entered_password):
    """Verify that if client_cert is not provided then a password must be."""
    settings_obj = settings.Settings(
        username='fakeuser', client_cert=client_cert)
    assert settings_obj.password == 'entered pw'


def test_client_cert_is_set_and_password_not_if_both_given(entered_password):
    """Verify that if both password and client_cert are given they are set"""
    client_cert = '/random/path'
    settings_obj = settings.Settings(
        username='fakeuser', password='anything', client_cert=client_cert)
    assert not settings_obj.password
    assert settings_obj.client_cert == client_cert


class TestArgumentParsing:

    @staticmethod
    def parse_args(args):
        parser = argparse.ArgumentParser()
        settings.Settings.register_argparse_arguments(parser)
        return parser.parse_args(args)

    def test_non_interactive_flag(self):
        args = self.parse_args(['--non-interactive'])
        assert args.non_interactive

    def test_non_interactive_environment(self, monkeypatch):
        monkeypatch.setenv("TWINE_NON_INTERACTIVE", "1")
        args = self.parse_args([])
        assert args.non_interactive
        monkeypatch.setenv("TWINE_NON_INTERACTIVE", "0")
        args = self.parse_args([])
        assert not args.non_interactive



import pytest

from twine import (
    auth,
    exceptions,
    utils,
)


cred = auth.CredentialInput


@pytest.fixture
def config() -> utils.RepositoryConfig:
    return dict(repository='system')


def test_get_password_keyring_overrides_prompt(monkeypatch, config):
    class MockKeyring:
        @staticmethod
        def get_password(system, user):
            return '{user}@{system} sekure pa55word'.format(**locals())

    monkeypatch.setattr(auth, 'keyring', MockKeyring)

    pw = auth.Resolver(config, cred('user')).password
    assert pw == 'user@system sekure pa55word'


def test_get_password_keyring_defers_to_prompt(
        monkeypatch, entered_password, config):
    class MockKeyring:
        @staticmethod
        def get_password(system, user):
            return

    monkeypatch.setattr(auth, 'keyring', MockKeyring)

    pw = auth.Resolver(config, cred('user')).password
    assert pw == 'entered pw'


def test_no_password_defers_to_prompt(monkeypatch, entered_password, config):
    config.update(password=None)
    pw = auth.Resolver(config, cred('user')).password
    assert pw == 'entered pw'


def test_empty_password_bypasses_prompt(
        monkeypatch, entered_password, config):
    config.update(password='')
    pw = auth.Resolver(config, cred('user')).password
    assert pw == ''


def test_no_username_non_interactive_aborts(config):
    with pytest.raises(exceptions.NonInteractive):
        auth.Private(config, cred('user')).password


def test_no_password_non_interactive_aborts(config):
    with pytest.raises(exceptions.NonInteractive):
        auth.Private(config, cred('user')).password


def test_get_username_and_password_keyring_overrides_prompt(
        monkeypatch, config):

    class MockKeyring:
        @staticmethod
        def get_credential(system, user):
            return cred(
                'real_user',
                'real_user@{system} sekure pa55word'.format(**locals())
            )

        @staticmethod
        def get_password(system, user):
            cred = MockKeyring.get_credential(system, user)
            if user != cred.username:
                raise RuntimeError("unexpected username")
            return cred.password

    monkeypatch.setattr(auth, 'keyring', MockKeyring)

    res = auth.Resolver(config, cred())
    assert res.username == 'real_user'
    assert res.password == 'real_user@system sekure pa55word'


@pytest.fixture
def keyring_missing_get_credentials(monkeypatch):
    """
    Simulate keyring prior to 15.2 that does not have the
    'get_credential' API.
    """
    monkeypatch.delattr(auth.keyring, 'get_credential')


@pytest.fixture
def entered_username(monkeypatch):
    monkeypatch.setattr(
        auth, 'input', lambda prompt: 'entered user', raising=False)


def test_get_username_keyring_missing_get_credentials_prompts(
        entered_username, keyring_missing_get_credentials, config):
    assert auth.Resolver(config, cred()).username == 'entered user'


def test_get_username_keyring_missing_non_interactive_aborts(
        entered_username, keyring_missing_get_credentials, config):
    with pytest.raises(exceptions.NonInteractive):
        auth.Private(config, cred()).username


def test_get_password_keyring_missing_non_interactive_aborts(
        entered_username, keyring_missing_get_credentials, config):
    with pytest.raises(exceptions.NonInteractive):
        auth.Private(config, cred('user')).password


@pytest.fixture
def keyring_no_backends(monkeypatch):
    """
    Simulate that keyring has no available backends. When keyring
    has no backends for the system, the backend will be a
    fail.Keyring, which raises RuntimeError on get_password.
    """
    class FailKeyring:
        @staticmethod
        def get_password(system, username):
            raise RuntimeError("fail!")
    monkeypatch.setattr(auth, 'keyring', FailKeyring())


@pytest.fixture
def keyring_no_backends_get_credential(monkeypatch):
    """
    Simulate that keyring has no available backends. When keyring
    has no backends for the system, the backend will be a
    fail.Keyring, which raises RuntimeError on get_credential.
    """
    class FailKeyring:
        @staticmethod
        def get_credential(system, username):
            raise RuntimeError("fail!")
    monkeypatch.setattr(auth, 'keyring', FailKeyring())


def test_get_username_runtime_error_suppressed(
        entered_username, keyring_no_backends_get_credential, recwarn,
        config):
    assert auth.Resolver(config, cred()).username == 'entered user'
    assert len(recwarn) == 1
    warning = recwarn.pop(UserWarning)
    assert 'fail!' in str(warning)


def test_get_password_runtime_error_suppressed(
        entered_password, keyring_no_backends, recwarn, config):
    assert auth.Resolver(config, cred('user')).password == 'entered pw'
    assert len(recwarn) == 1
    warning = recwarn.pop(UserWarning)
    assert 'fail!' in str(warning)



from twine.cli import dispatch


def test_devpi_upload(devpi_server, uploadable_dist):
    command = [
        'upload',
        '--repository-url', devpi_server.url,
        '--username', devpi_server.username,
        '--password', devpi_server.password,
        str(uploadable_dist),
    ]
    dispatch(command)


twine_sampleproject_token = (
    "pypi-AgENdGVzdC5weXBpLm9yZwIkNDgzYTFhMjEtMzEwYi00NT"
    "kzLTkwMzYtYzc1Zjg4NmFiMjllAAJEeyJwZXJtaXNzaW9ucyI6IH"
    "sicHJvamVjdHMiOiBbInR3aW5lLXNhbXBsZXByb2plY3QiXX0sIC"
    "J2ZXJzaW9uIjogMX0AAAYg2kBZ1tN8lj8dlmL3ScoVvr_pvQE0t"
    "6PKqigoYJKvw3M"
)


def test_pypi_upload(sampleproject_dist):
    command = [
        'upload',
        '--repository-url', 'https://test.pypi.org/legacy/',
        '--username', '__token__',
        '--password', twine_sampleproject_token,
        str(sampleproject_dist),
    ]
    dispatch(command)


def test_pypiserver_upload(pypiserver_instance, uploadable_dist):
    command = [
        'upload',
        '--repository-url', pypiserver_instance.url,
        '--username', 'any',
        '--password', 'any',
        str(uploadable_dist),
    ]
    dispatch(command)



from __future__ import unicode_literals

import pytest
import pretend

from twine.commands import register
from twine import exceptions


# TODO: Copied from test_upload.py. Extract to helpers?

WHEEL_FIXTURE = 'tests/fixtures/twine-1.5.0-py2.py3-none-any.whl'


def test_exception_for_redirect(make_settings):
    register_settings = make_settings("""
        [pypi]
        repository: https://test.pypi.org/legacy
        username:foo
        password:bar
    """)

    stub_response = pretend.stub(
        is_redirect=True,
        status_code=301,
        headers={'location': 'https://test.pypi.org/legacy/'}
    )

    stub_repository = pretend.stub(
        register=lambda package: stub_response,
        close=lambda: None
    )

    register_settings.create_repository = lambda: stub_repository

    with pytest.raises(exceptions.RedirectDetected) as err:
        register.register(register_settings, WHEEL_FIXTURE)

    assert "https://test.pypi.org/legacy/" in err.value.args[0]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from twine import __main__ as dunder_main
from twine import exceptions

import pretend


def test_exception_handling(monkeypatch):
    replaced_dispatch = pretend.raiser(
        exceptions.InvalidConfiguration('foo')
    )
    monkeypatch.setattr(dunder_main, 'dispatch', replaced_dispatch)
    assert dunder_main.main() == 'InvalidConfiguration: foo'



# Copyright 2015 Ian Cordasco
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from twine import wheel

import pytest


@pytest.fixture(params=[
    'tests/fixtures/twine-1.5.0-py2.py3-none-any.whl',
    'tests/alt-fixtures/twine-1.5.0-py2.py3-none-any.whl'
])
def example_wheel(request):
    return wheel.Wheel(request.param)


def test_version_parsing(example_wheel):
    assert example_wheel.py_version == 'py2.py3'


def test_find_metadata_files():
    names = [
        'package/lib/__init__.py',
        'package/lib/version.py',
        'package/METADATA.txt',
        'package/METADATA.json',
        'package/METADATA',
    ]
    expected = [
        ['package', 'METADATA'],
        ['package', 'METADATA.json'],
        ['package', 'METADATA.txt'],
    ]
    candidates = wheel.Wheel.find_candidate_metadata_files(names)
    assert expected == candidates



# Copyright 2015 Ian Cordasco
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import requests

from twine import repository
from twine.utils import DEFAULT_REPOSITORY, TEST_REPOSITORY

import pretend
import pytest
from contextlib import contextmanager


def test_gpg_signature_structure_is_preserved():
    data = {
        'gpg_signature': ('filename.asc', 'filecontent'),
    }

    tuples = repository.Repository._convert_data_to_list_of_tuples(data)
    assert tuples == [('gpg_signature', ('filename.asc', 'filecontent'))]


def test_content_structure_is_preserved():
    data = {
        'content': ('filename', 'filecontent'),
    }

    tuples = repository.Repository._convert_data_to_list_of_tuples(data)
    assert tuples == [('content', ('filename', 'filecontent'))]


def test_iterables_are_flattened():
    data = {
        'platform': ['UNKNOWN'],
    }

    tuples = repository.Repository._convert_data_to_list_of_tuples(data)
    assert tuples == [('platform', 'UNKNOWN')]

    data = {
        'platform': ['UNKNOWN', 'ANOTHERPLATFORM'],
    }

    tuples = repository.Repository._convert_data_to_list_of_tuples(data)
    assert tuples == [('platform', 'UNKNOWN'),
                      ('platform', 'ANOTHERPLATFORM')]


def test_set_client_certificate():
    repo = repository.Repository(
        repository_url=DEFAULT_REPOSITORY,
        username='username',
        password='password',
    )

    assert repo.session.cert is None

    repo.set_client_certificate(('/path/to/cert', '/path/to/key'))
    assert repo.session.cert == ('/path/to/cert', '/path/to/key')


def test_set_certificate_authority():
    repo = repository.Repository(
        repository_url=DEFAULT_REPOSITORY,
        username='username',
        password='password',
    )

    assert repo.session.verify is True

    repo.set_certificate_authority('/path/to/cert')
    assert repo.session.verify == '/path/to/cert'


def test_make_user_agent_string():
    repo = repository.Repository(
        repository_url=DEFAULT_REPOSITORY,
        username='username',
        password='password',
    )

    assert 'User-Agent' in repo.session.headers

    user_agent = repo.session.headers['User-Agent']
    assert 'twine/' in user_agent
    assert 'requests/' in user_agent
    assert 'requests-toolbelt/' in user_agent
    assert 'pkginfo/' in user_agent
    assert 'setuptools/' in user_agent


def response_with(**kwattrs):
    resp = requests.Response()
    for attr, value in kwattrs.items():
        if hasattr(resp, attr):
            setattr(resp, attr, value)

    return resp


def test_package_is_uploaded_404s():
    repo = repository.Repository(
        repository_url=DEFAULT_REPOSITORY,
        username='username',
        password='password',
    )
    repo.session = pretend.stub(
        get=lambda url, headers: response_with(status_code=404)
    )
    package = pretend.stub(
        safe_name='fake',
        metadata=pretend.stub(version='2.12.0'),
    )

    assert repo.package_is_uploaded(package) is False


def test_package_is_uploaded_200s_with_no_releases():
    repo = repository.Repository(
        repository_url=DEFAULT_REPOSITORY,
        username='username',
        password='password',
    )
    repo.session = pretend.stub(
        get=lambda url, headers: response_with(status_code=200,
                                               _content=b'{"releases": {}}',
                                               _content_consumed=True),
    )
    package = pretend.stub(
        safe_name='fake',
        metadata=pretend.stub(version='2.12.0'),
    )

    assert repo.package_is_uploaded(package) is False


@pytest.mark.parametrize('disable_progress_bar', [
    True,
    False
])
def test_disable_progress_bar_is_forwarded_to_tqdm(monkeypatch, tmpdir,
                                                   disable_progress_bar):
    """Test whether the disable flag is passed to tqdm
        when the disable_progress_bar option is passed to the
        repository
    """
    @contextmanager
    def progressbarstub(*args, **kwargs):
        assert "disable" in kwargs
        assert kwargs["disable"] == disable_progress_bar
        yield

    monkeypatch.setattr(repository, "ProgressBar", progressbarstub)
    repo = repository.Repository(
        repository_url=DEFAULT_REPOSITORY,
        username='username',
        password='password',
        disable_progress_bar=disable_progress_bar,
    )

    repo.session = pretend.stub(
        post=lambda url, data, allow_redirects, headers: response_with(
            status_code=200)
    )

    fakefile = tmpdir.join('fake.whl')
    fakefile.write('.')

    def dictfunc():
        return {"name": "fake"}

    package = pretend.stub(
        safe_name='fake',
        metadata=pretend.stub(version='2.12.0'),
        basefilename="fake.whl",
        filename=str(fakefile),
        metadata_dictionary=dictfunc
    )

    repo.upload(package)


@pytest.mark.parametrize('package_meta,repository_url,release_urls', [
    # Single package
    (
        [('fake', '2.12.0')],
        DEFAULT_REPOSITORY,
        {'https://pypi.org/project/fake/2.12.0/'},
    ),
    # Single package to testpypi
    (
        [('fake', '2.12.0')],
        TEST_REPOSITORY,
        {'https://test.pypi.org/project/fake/2.12.0/'},
    ),
    # Multiple packages (faking a wheel and an sdist)
    (
        [('fake', '2.12.0'), ('fake', '2.12.0')],
        DEFAULT_REPOSITORY,
        {'https://pypi.org/project/fake/2.12.0/'},
    ),
    # Multiple releases
    (
        [('fake', '2.12.0'), ('fake', '2.12.1')],
        DEFAULT_REPOSITORY,
        {
            'https://pypi.org/project/fake/2.12.0/',
            'https://pypi.org/project/fake/2.12.1/',
        },
    ),
    # Not pypi
    (
        [('fake', '2.12.0')],
        'http://devpi.example.com',
        set(),
    ),
    # No packages
    (
        [],
        DEFAULT_REPOSITORY,
        set(),
    ),
])
def test_release_urls(package_meta, repository_url, release_urls):
    packages = [
        pretend.stub(
            safe_name=name,
            metadata=pretend.stub(version=version),
        )
        for name, version in package_meta
    ]

    repo = repository.Repository(
        repository_url=repository_url,
        username='username',
        password='password',
    )

    assert repo.release_urls(packages) == release_urls



#!/usr/bin/env python3
# Copyright 2013 Donald Stufft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

import requests

from twine import exceptions
from twine.cli import dispatch


def main():
    try:
        return dispatch(sys.argv[1:])
    except (exceptions.TwineException, requests.exceptions.HTTPError) as exc:
        return '{}: {}'.format(exc.__class__.__name__, exc.args[0])


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2015 Ian Cordasco
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import os.path

from twine.package import PackageFile
from twine import exceptions
from twine import settings


def register(register_settings, package):
    repository_url = register_settings.repository_config['repository']

    print(f"Registering package to {repository_url}")
    repository = register_settings.create_repository()

    if not os.path.exists(package):
        raise exceptions.PackageNotFound(
            f'"{package}" does not exist on the file system.'
        )

    resp = repository.register(
        PackageFile.from_filename(package, register_settings.comment)
    )
    repository.close()

    if resp.is_redirect:
        raise exceptions.RedirectDetected.from_args(
            repository_url,
            resp.headers["location"],
        )

    resp.raise_for_status()


def main(args):
    parser = argparse.ArgumentParser(
        prog="twine register",
        description="register operation is not required with PyPI.org",
    )
    settings.Settings.register_argparse_arguments(parser)
    parser.add_argument(
        "package",
        metavar="package",
        help="File from which we read the package metadata.",
    )

    args = parser.parse_args(args)
    register_settings = settings.Settings.from_argparse(args)

    # Call the register function with the args from the command line
    register(register_settings, args.package)
